;;; htmlj-mode.el --- HTML mode made by a Japanese.

;; Copyright (C) 1996-1998 Gaku UEDA
;; Maintaier: Gaku UEDA <gaku@ueda.info.waseda.ac.jp>
;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; A copy of the GNU General Public License can be obtained from
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
;;; 02139, USA.

;; Version 1.50 (1998/02/02)
;; Takuzo OHARA <takuzo@ueda.info.waseda.ac.jp> has made a patch 
;; which enable htmlj-mode to handle font-lock facilities. 
;; This takuzo-patch has marged to 1.50. (Thanks Takuzo)
;; Little bug fix for hilit-set-mode-patterns part of his patch.
;;
;; Version 1.49 (1998/01/30)
;; Merge Watanabe patch to be able to call skelton function by C-c C-s.
;; (Thanks to Watanabe Takashi <watanabe@ueda.info.waseda.ac.jp>)
;; Change function name from insert-skel to htmlj-insert-skelton.
;;
;; Version 1.48 (1998/01/29)
;; Add a missing definition for <DT> tag.
;; Bug fix for C-c i.
;;
;; Version 1.47 (1998/01/28)
;; Add a useful command to close tag (C-c C-c).
;; Bug fix for htmlj-get-this-tag-whole.
;; Bug fix for <IMG> tag insert. 
;; Now it recognize beginning '<' as IN a tag. But no difference for users.
;; Add htmlj-in-a-tag-p. 
;;
;; Version 1.46 (1997/10/03) 
;; Add a function for adding URL on string at the point.
;; htmlj-put-url-tag-on-string
;;
;; Version 1.45 (1997/01/29) 
;; Add a function for insert <BR>. (C-c j)
;; htmlj-insert-br
;;
;; Version 1.44 (1997/01/24) 
;; Fix a bug of htmlj-replace-kakko.
;;
;; Version 1.43 (1997/01/22) 
;; C-c C-l becomes a bind for insert a link tag for region.
;; function: htmlj-region-insert-link
;; function: htmlj-kakomu-region-new
;;
;; Version 1.42 (1997/01/02) 
;; C-c C-l becomes a bind for insert a link.
;;
;; Version 1.41 (1996/11/04 23:50:18)
;; Add code for non hilit19 emacs environment. 
;;
;; Version 1.40 (1996/11/04 23:33:06)
;; Enhance generic tag input function.  Now, htmlj-mode can control 
;; how to input a tag depend on tag's characteristics.  The character
;; -istics is defined by htmlj-tag-info-struct.
;;
;; Version 1.35 (1996/11/04 21:25:44)
;; Define a new user preference for tag case policy.
;; If you always want to input a tag in upper case, set 
;; htmlj-tag-should-uppercase , t , otherwize, set nil.
;;
;; Version 1.34 (1996/11/04 21:14:49)
;; Fix a bug happen when insert TAGs continuously.
;;
;; Version 1.33 (1996/11/04 02:58:14)
;; Now covered almost tags. Over 80 tags!
;;
;; Version 1.32 (1996/10/28 02:51:48)
;; Tweak how supply the attributes for <BODY> tag. 
;;
;; Version 1.31 (1996/10/28 02:34:29)
;; For HTML sample, translate '<'s and '>'s in the region into 
;; '&lt;'s and '&gt's. 
;;
;; Version 1.3  (1996/10/28 02:28:17)
;; Add attribute suppling feature.
;; It enable a user to supply optional attributes interactively.
;; Attributes list is defined as defvar htmlj-tag-attribute-alist.
;;
;; Version 1.22 (1996/10/26 13:05:09)
;; Give <BODY>tag with default attribute parameter.
;;  defvar htmlj-mode-body-default-param
;; Add defun for <IMG> tag.
;;
;; Version 1.21 (1996/10/25 11:43:47)
;; When region begin and end within same line, put a tag without CR.
;; It become more readable.
;;
;; Version 1.20 (1996/10/25 11:13:07)
;; Add surround region with a tag feature (htmlj-kakomu-region-interactive)
;; (C-c C-t)
;; Separate complition function as htmlj-complete-tag.
;;
;; Version 1.11 (1996/10/25 10:52:41)
;; Add Brace quote feature (htmlj-replace-kakko) 
;;
;; Version 1.10 (1996/10/25 10:48:06)
;; Add tag completion feature (C-c t)
;;
;; Version 1.00 (1996/10/14 01:57:29)
;; Initial version


;;; INSTALL:
;; To use htmlj-mode, add lines below in ~/.emacs
;;
;; (autoload 'htmlj-mode "htmlj-mode" "Gaku's html mode" t)
;; (setq auto-mode-alist
;;   (append '(("\\.html$" . htmlj-mode)) auto-mode-alist))
;;
;;
;;; Commentary:

;;; Code: 

; ("LI" . [single nocr]
;  "LI" $BO"A[G[Ns$+$i%?%0$r0z$CD%$C$F$/$k$?$a$N%-!<!#BgJ8;z$G=q$/!#(B
;  single|close  close:<PRE>$B!A(B</PRE>$B$N$h$&$KJD$8$kI,MW$,$"$k$b$N(B
;                single: <LI>$B$N$h$&$K(B</LI>$B$,>JN,2DG=$J$b$N(B
;  nocr|cr       nocr: $BJ8;zNs$r64$`$b$N!#(B<B>hello</B>$B$J$I!#(B
;                cr: $BNN0h$r64$`$?$a!"2~9T$rF~NO$7$F$*$-$?$$>l9g!#$?$H$($P(B
;                <PRE>$B!&"+$3$3$H(B
;                 $B$3$3$,NN0h!&"+$3$3$K2~9T$rF~$l$k!#(B
;                </PRE>
(setq htmlj-tag-info-struct
  '(("LI" . [single nocr])
    ("UL" . [close cr])
    ("OL" . [close cr])
    ("DL" . [close cr])
    ("DD" . [single nocr])
    ("DT" . [single nocr])
    ("MENU" . [close cr])

    ("H1" . [close nocr])
    ("H2" . [close nocr])
    ("H3" . [close nocr])
    ("H4" . [close nocr])
    ("H5" . [close nocr])
    ("H6" . [close nocr])

    ("HTML" . [close cr])
    ("TITLE" . [close nocr])
    ("BODY"  . [close cr])
    ("HEAD" . [close cr])

    ("CENTER" . [close cr])
    ("PRE" . [close cr])
    ("BLOCKQUOTE" . [close cr])
    ("XMP" . [close cr])
    ("NOBR" . [close cr])

    ("A"  . [close  nocr])
    ("B"  . [close  nocr])
    ("I"  . [close nocr])
    ("EM" . [close nocr])
    ("FONT" . [close nocr])
    ("CODE" . [close nocr])
    ("STRONG" . [close nocr])
    ("BLINK" . [close nocr])	
    ("ADDRESS" . [close nocr])
    ("SUB" . [close nocr])
    ("SUP" . [close nocr])
    ("U" . [close nocr])
    ("KBD" . [close nocr])
    ("SMALL" . [close nocr])
    ("BIG" . [close nocr])

    ("FORM" . [close cr])
    ("INPUT" . [single nocr])

    ("SCRIPT" . [close cr])

    ("APPLET" . [close cr])
    ("PARAM" . [single cr])

    ("FRAMESET" . [close cr])
    ("FRAME" . [single cr])


    ("TABLE" . [close cr])
    ("TR" . [single nocr])
    ("TD" . [single nocr])
    ("TH" . [single nocr])
    ("TT" . [single nocr])

    ("IMG" . [single nocr])

    ("P"  . [single cr])
    ("HR" . [single cr])
    ("BR" . [single cr])

;;$B$h$/$o$+$i$J$$$b$N(B
; $B%(%s%H%j$,$J$$>l9g$O!"(B[close nobr]$B$H$7$F07$o$l$F$$$k$O$:$@!#(B
;    ("META" . [])
;    ("BASEFONT" . [])
;    ("CITE" . [])
;    ("CAPTION" . [])
;    ("L1" . [])
;    ("DIV" . [])
;    ("SELECT" . [])
;    ("AREA" . [])
;    ("BASE" . [])
;    ("INTERLACE" . [])
;    ("TEXTAREA" . [])
;    ("V" . [])
;    ("IMAGE" . [])
;    ("SAMP" . [])
;    ("LINK" . [])
;    ("J" . [])
;    ("BLOCK" . [])
;    ("C" . [])
;    ("DI" . [])
;    ("VAR" . [])
;    ("NOFRAMES" . [])
;    ("HEADER" . [])
;    ("VN" . [])
;    ("MAP" . [])
;    ("MARQUEE" . [])
;    ("LEFT" . [])
;    ("EMBED" . [])
;    ("APP" . [])
;   ("OPTION" . [unknown unknown])    
))

(defvar htmlj-mode-tag-list
  '(("A") ("BR") ("P") ("LI") ("B") ("TD") ("IMG") ("DD") ("HR")
    ("I") ("DT") ("FONT") ("TR") ("UL") ("H2") ("TITLE") ("DL")
    ("BODY") ("CENTER") ("H3") ("HEAD") ("H1") ("HTML") ("TH") ("PRE") ("TABLE") ("ADDRESS") ("EM") ("H4") ("OPTION") ("CODE") ("INPUT") ("STRONG") ("BLOCKQUOTE") ("OL") ("BLINK") ("TT") ("H5") ("META") ("SUB") ("PARAM") ("U") ("KBD") ("BASEFONT") ("MENU") ("CITE") ("FORM") ("CAPTION") ("L1") ("DIV") ("XMP") ("SUP") ("SCRIPT") ("SELECT") ("H6") ("SMALL") ("AREA") ("BASE") ("FRAME") ("INTERLACE") ("TEXTAREA") ("APPLET") ("V") ("IMAGE") ("SAMP") ("LINK") ("FRAMESET") ("BIG") ("J")  ("C") ("NOBR") ("DI") ("VAR") ("NOFRAMES") ("HEADER") ("VN") ("MAP") ("MARQUEE") ("LEFT") ("EMBED") ("APP")) 
  "List of tags.  This list is used to complete a tag." 
)
;("BLOCK") $B=|30(B


(defvar htmlj-tag-attribute-alist
  '(("A" . ("HREF" "NAME" "TARGET"))
    ("BR" . ("CLEAR"))
    ("LI" . ("TYPE"))
    ("TD" . ("ALIGN" "WIDTH" "COLSTART" "VALIGN" "ROWSPAN" "COLSPAN" "NOWRAP"
	     "BGCOLOR"))

    ("IMG" . ("SRC" "ALT" "WIDTH" "HEIGHT" "ALIGN" "BORDER" "HSPACE" "VSPACE"
	      "LOWSRC" "ISMAP" "NATURALSIZEFLAG" "USEMAP"))
    ("HR" . ("SIZE" "WIDTH" "ALIGN" "NOSHADE"))
    ("FONT" . ("SIZE" "COLOR" "FACE"))
    ("TR" . ("ALIGN" "VALIGN" "BGCOLOR" "COLSPAN"))
    ("UL" . ("TYPE" "COMPACT")) 
    ("H2" . ("ALIGN")) 
    ("TITLE" . ("BETWEEN")) 
    ("DL" . ("COMPACT"))
    ("BODY" . ("BGCOLOR" "TEXT" "VLINK" "LINK" "BACKGROUND" "ALINK" "onLOAD"))
    ("H3" . ("ALIGN" "ADD_DATE" "FOLDED"))
    ("H1" . ("ALIGN")) 
    ("HTML" . ("VERSION"))
    ("TH" . ("ALIGN" "WIDTH" "VALIGN" "COLSPAN" "ROWSPAN" "BGCOLOR"
	     "COLSTART" "NOWRAP"))
    ("TABLE" . ("BORDER" "WIDTH" "CELLPADDING" "CELLSPACING" "ALIGN" "HEIGHT"
		"BGCOLOR" "VSPACE" "HSPACE" "NOSHADE" "BORDERCOLOR" "COLSPEC"))
    ("H4" . ("ALIGN"))
    ("OPTION" . ("VALUE" "SELECTED"))
    ("INPUT" . ("TYPE" "NAME" "VALUE" "SIZE" "ONCLICK" "MAXLENGTH" "CHECKED"))
    ("OL" . ("TYPE" "START"))
    ("H5" . ("ALIGN"))
    ("META" . ("CONTENT" "NAME" "HTTP-EQUIV" "VALUE"))
    ("PARAM" . ("NAME" "VALUE"))
    ("BASEFONT" . ("SIZE"))
    ("FORM" . ("METHOD" "ACTION" "NAME" "ONSUBMIT" "ENCTYPE" "TARGET"
	       "onRESET"))
    ("CAPTION" . ("ALIGN" "VALIGN" "CENTER"))
    ("DIV" . ("ALIGN"))
    ("SELECT" . ("NAME" "SIZE" "MULTIPLE" "onChange"))
    ("H6" . ("ALIGN"))
    ("AREA" . ("COORDS" "SHAPE" "HREF" "TARGET" "NOHREF" "NAME")) 
    ("BASE" . ("HREF" "TARGET"))
    ("FRAME" . ("SRC" "NAME" "SCROLLING" "MARGINWIDTH" "MARGINHEIGHT"
		"NORESIZE"))
    ("TEXTAREA" . ("NAME" "COLS" "ROWS"))
    ("APPLET" . ("CODE" "HEIGHT" "WIDTH" "CODEBASE" "NAME" "ALIGN"
		 "CLEARTOWIN" "MAYSCRIPT"))
    ("MARQUEE" . ("WIDTH" "DIRECTION" "BGCOLOR" "ALIGN" "SCROLLDELAY"
		  "BORDER" "SCROLLAMOUNT" "BEHAVIOR" "LOOP"))
    ("EMBED" . ("WIDTH" "SRC" "HEIGHT" "ALIGN"))
    ("APP" . ("CLASS" "WIDTH" "HEIGHT"))
))

(defvar htmlj-mode-skelton-function 'htmlj-mode-default-skelton-function
  "skelton function symbol")

(defvar htmlj-mode-syntax-table nil
  "Syntax table used while in htmlj mode.")

(defvar htmlj-mode-abbrev-table nil
  "Abbrev table used while in htmlj mode.")
(define-abbrev-table 'htmlj-mode-abbrev-table ())

(defvar htmlj-mode-body-default-param "BGCOLOR=\"#FFFFFF\""
  "Default paramter for <BODY> tag. If you want use it, remember\
   the first space is very important.")

(defvar htmlj-tag-should-uppercase t
  "If you always want to input a tag in uppercase, set this t.")

(if htmlj-mode-syntax-table
    ()
  (setq htmlj-mode-syntax-table (make-syntax-table)))

(defvar htmlj-mode-map nil
  "Keymap for htmlj mode.")

(if htmlj-mode-map
    ()
  (setq htmlj-mode-map (make-sparse-keymap))
  (define-key htmlj-mode-map "\C-c1" 'insert-h1)
  (define-key htmlj-mode-map "\C-c2" 'insert-h2)
  (define-key htmlj-mode-map "\C-c3" 'insert-h3)
  (define-key htmlj-mode-map "\C-cl" 'htmlj-insert-link)
  (define-key htmlj-mode-map "\C-cL" 'htmlj-put-url-tag-on-string)
  (define-key htmlj-mode-map "\C-c\C-l" 'htmlj-region-insert-link)  
  (define-key htmlj-mode-map "\C-cp" 'htmlj-insert-pp)
  (define-key htmlj-mode-map "\C-cj" 'htmlj-insert-br)
  (define-key htmlj-mode-map "\C-cs" 'htmlj-insert-skelton)
  (define-key htmlj-mode-map "\C-cu" 'insert-ul)
  (define-key htmlj-mode-map "\C-ci" 'htmlj-insert-li)
  (define-key htmlj-mode-map "\C-cq" 'insert-tagstr)
  (define-key htmlj-mode-map "\C-ct" 'htmlj-insert-tag-with-completion)
  (define-key htmlj-mode-map "\C-c\C-t" 'htmlj-kakomu-region-interactive)
  (define-key htmlj-mode-map "\C-c\C-q" 'htmlj-replace-kakko)
  (define-key htmlj-mode-map "\C-cg" 'htmlj-img-tag)
  (define-key htmlj-mode-map "\C-c " 'htmlj-supply-attrib)
  (define-key htmlj-mode-map "\C-c\C-c" 'htmlj-insert-closing-tag)
)

(defun htmlj-mode ()
  "Major mode for editing html"
  (interactive)
  (kill-all-local-variables)
  (use-local-map htmlj-mode-map)
  (setq mode-name "htmlj")
  (setq major-mode 'htmlj-mode)
  (setq local-abbrev-table htmlj-mode-abbrev-table)
  (set-syntax-table htmlj-mode-syntax-table)
  ;; for Font-lock 
  (cond	((string-match "XEmacs\\|Lucid" (emacs-version)) ; XEmacs/Lucid
	 (setq font-lock-keywords-case-fold-search t)))
  (run-hooks 'htmlj-mode-hook))

; hilit19$B$,;H$o$l$F$$$k;~!"Ce?'$9$k!#(B
; Time-stamp: <1998/02/02 02:59:47 takuzo>
; Bug fix 02/02/1998 Gaku UEDA
(cond ((featurep 'hilit19)
       (hilit-set-mode-patterns
	'htmlj-mode
	'(
	  ("<h[1-9]>" "</h[1-9]>" label)
	  ("<a href" "</a>" label)
	  ("<p>" nil label)
	  ("<!--" "-->" comment)
	  ("<![a-z]+\\>[^<>]*\\(<[^>]*>[^<>]*\\)*>" nil comment)
	  ("<b>" "</b>" bold)
	  ("<i>" "</i>" italic)
	  ("<u>" "</u>" underline)
	  ("&[^;\n]*;" nil label)
	  ("<[^>]*>" nil defun))
	nil 'case-insensitive)))

(defun insert-tagstr (str)
  (interactive "stag string: ")
  (insert (format "&lt;%s&gt;" str))) 

; 1996/10/24 21:31:47
(defun htmlj-insert-li () (interactive) (htmlj-insert-tag "LI"))
(defun htmlj-insert-pp () (interactive) (htmlj-insert-tag "P"))
(defun htmlj-insert-br () (interactive) (htmlj-insert-tag "BR"))

(defun htmlj-insert-applet-tag () (interactive) (htmlj-insert-tag "APPLET"))


; 1996/10/24 21:31:50
(defun insert-ul ()
  (interactive)
  (insert "<UL>\n")
  (insert " <LI>\n")
  (insert "</UL>\n"))

(defun insert-h1 (str)
  (interactive "s<H1>:")
  (insert-head 1 str))

(defun insert-h2 (str)
  (interactive "s<H2>:")
  (insert-head 2 str))

(defun insert-h3 (str)
  (interactive "s<H3>:")
  (insert-head 3 str))

(defun insert-head (n str)
  (insert (format "<H%d>%s</H%d>\n" n str n)))


(defun htmlj-insert-link (url str)
  (interactive "sURL for HREF:\nsdisplay text:")
  (insert "<A HREF=\"")
  (insert url)
  (insert "\">")
  (insert str)
  (insert "</A>"))

;; $B8=:_$N%j!<%8%g%s$K%j%s%/$r@_Dj$9$k!#(B 
(defun htmlj-region-insert-link (start end)
  (interactive "*r")
  (let ((url "") (stag "") (etag "</A>"))
    (setq url (read-string "URL for HREF:"))
    (setq stag (format "<A HREF=\"%s\">" url))
    (htmlj-kakomu-region-new start end stag etag)))

(defun htmlj-insert-skelton ()
  (interactive)
   (if (and htmlj-mode-skelton-function
 	     (fboundp htmlj-mode-skelton-function))
 	  (progn
 	    (insert (funcall htmlj-mode-skelton-function))
 	    )))

(defun htmlj-mode-default-skelton-function ()
   "default skelton function"
    (let ((param ""))
      (or (equal htmlj-mode-body-default-param "") (setq param " "))
     (format "<HTML><HEAD><TITLE>\n<HTML>\n<HEAD>\n</HEAD>\n<BODY%s%s>\n\n</BODY>\n</HTML>" param htmlj-mode-body-default-param)))

; $B%?%0$NF~NO$r9T$J$&(B($BJd405!G=$D$-(B)
(defun htmlj-insert-tag-with-completion ()
  (interactive)
  (let ((tag))
    (setq tag (htmlj-complete-tag))
    (cond ((string= tag "SCRIPT") (htmlj-insert-script-tag))
	  ((string= tag "APPLET") (htmlj-insert-applet-tag))
	  (t (htmlj-insert-tag tag)))))

; $BC1=c$K%?%0$r%P%C%U%!$K=PNO$9$k!#(Bhtmlj-insert-tag-with-completion$B$N2<@A$1(B
(defun htmlj-insert-tag (str)
  (interactive "sTag: ")
  (let ((m) ; m: the position for cursol after this process
	(tag) ; tag: uppercase version of str used as a key of assoc list
	)
    (setq tag (upcase str))
    ; $BI=<(%?%0$N@55,2=$,I,MW$J$i$d$k!#(B
    (if htmlj-tag-should-uppercase
	(setq str (upcase str)))

    (insert (format "<%s>" str)) ; $B3+;O%?%0F~NO(B
    (if (htmlj-tag-need-cr-p tag)
	(insert "\n"))
    (cond ((not (htmlj-tag-need-single-p tag))
	   (progn (setq m (point-marker)) ; $B3+;O%?%0$ND>8e$N0LCV$r5-21!#(B
		  (insert (format "</%s>" str))
		  (if (htmlj-tag-need-cr-p tag)
		      (insert "\n"))
		  (goto-char (marker-position m)))))))

; $B%?%0$rF~NO$9$k:]!"2~9T$,I,MW$+$I$&$+!#(Bhtmlj-insert-tag$B$+$i8F$P$l$k!#(B
(defun htmlj-tag-need-cr-p (tag)
  (eq 'cr (elt (cdr (assoc tag htmlj-tag-info-struct)) 1))
)

; $BJD$8%?%0(B(</ABC>)$B$N$$$i$J$$$b$N$+!)(B
(defun htmlj-tag-need-single-p (tag)
  (eq 'single (elt (cdr (assoc tag htmlj-tag-info-struct)) 0))
)
  


; <SCRIPT>$B%?%0$N=PNO(B
; JavaScript$B$N$?$a$N%?%0$r=PNO$9$k!#HsBP1~%V%i%&%6$N$?$a$N%3%a%s%H$b(B
; $BF1;~$K=PNO$9$k!#(B
(defun htmlj-insert-script-tag ()
   (interactive)
   (insert "<SCRIPT LANGUAGE=\"JavaScript\">\n")
   (insert "<!-- Code hiding from here\n")
   (insert "\n")
   (insert "// end -->\n")
   (insert "</SCRIPT>\n")
)

	
; region$BCf$N(B"<", ">", "&"$B$r$=$l$>$l(B"&lt;", "&gt;", "&amp;"$B$KCV$-49$($k(B
;
; 1996/10/07 02:27:20
(defun htmlj-replace-kakko (start end)
  "Comment out region"
  (interactive "*r")
  (save-excursion
    (save-restriction
      (narrow-to-region end start)
      (goto-char (point-min))                  ; $B$b$&0lEY@hF,$K%]%$%s%H$rLa$7(B
      (perform-replace "&" "&amp;" nil nil nil) ; "&" $B"*(B "&amp;"$B$KCV49(B
      (narrow-to-region end start)             ; $B=hM}$r(Bregion$B$K8BDj(B
      (goto-char (point-min))                  ; $B@hF,$K%]%$%s%H$r0\F0$7(B
      (perform-replace ">" "&gt;" nil nil nil) ; ">" $B"*(B "&gt;"$B$KCV49(B
      (goto-char (point-min))                  ; $B$b$&0lEY@hF,$K%]%$%s%H$rLa$7(B
      (perform-replace "<" "&lt;" nil nil nil) ; "<" $B"*(B "&lt;"$B$KCV49(B
      )))

; $B%j!<%8%g%s$r%_%K%P%C%U%!$GF~NO$7$?%?%0$G0O$`(B
(defun htmlj-kakomu-region-interactive (start end)
  (interactive "*r")
  (htmlj-kakomu-region start end (htmlj-complete-tag)))

; $B<B:]$K%j!<%8%g%s$r%?%0$G0O$&(B
(defun htmlj-kakomu-region (start end tag)
  (save-excursion
    (save-restriction
      (narrow-to-region end start)
      (let ((needcr (htmlj-in-same-line-p start end)))
	(goto-char (point-min))
	(if needcr (insert (format "<%s>" tag))
	  (insert (format "<%s>\n" tag)))
	(goto-char (point-max))
	(if needcr (insert-before-markers (format "</%s>" tag))
	  ; insert-before-markers$B$@$H!"%^!<%+$ND>A0$KJ8;zNs$rF~$l$i$l$k!#(B
	  (insert-before-markers (format "</%s>\n" tag)))))))

;; start $B%j!<%8%g%s@hF,(B
;; end   $B%j!<%8%g%sKvHx(B
;; stag  $B3+;O%?%0(B  (<$B$r4^$a$9$Y$F(B)
;; etag  $B=*N;%?%0(B  (</$B$r4^$a$9$Y$F(B)
(defun htmlj-kakomu-region-new (start end stag etag)
  (save-excursion
    (save-restriction
      (narrow-to-region end start)
      (let ((needcr (htmlj-in-same-line-p start end)))
	(goto-char (point-min))
	(if needcr (insert (format "%s" stag))
	  (insert (format "%s\n" stag)))
	(goto-char (point-max))
	(if needcr (insert-before-markers (format "%s" etag))
	  ; insert-before-markers$B$@$H!"%^!<%+$ND>A0$KJ8;zNs$rF~$l$i$l$k!#(B
	  (insert-before-markers (format "%s\n" etag)))))))


; $B%?%0$NJd40$r9T$J$&(B
(defun htmlj-complete-tag ()
  (interactive)
  (let ((completion-ignore-case t)
	(rawtag))
    (setq rawtag (completing-read "Tag: " htmlj-mode-tag-list nil nil nil))
    (if htmlj-tag-should-uppercase 
	(upcase rawtag)
      rawtag))
)


; $B%j!<%8%g%s$,J#?t9T$KEO$k$+$I$&$+$r%A%'%C%/$9$k(B
(defun htmlj-in-same-line-p (start end)
  (interactive "*r")
  (save-excursion
    (let ((sp) (ep))
      (goto-char start)     ; $B%j!<%8%g%s$N@hF,$r9TF,$K$b$C$F$-$?$b$N$H(B
      (beginning-of-line)
      (setq sp (point))
      (goto-char end)       ; $B%j!<%8%g%sKvHx$r9TF,$K$b$C$F$-$?$b$N$,(B
      (beginning-of-line)
      (setq ep (point))
      (= sp ep))))          ; $BF1$8$J$iF1$89T$KN>J}$"$k$3$H$K$J$k(B

; $B%j!<%8%g%s$r%5%s%W%k$H$7$F!"(B
;  1. $BI=<($G$-$J$$J8;z$r%(%9%1!<%W(B
;  2. <PRE>$B$G0O$_(B
;  3. <BLOCKQUOTE>$B$G0O$`(B

;(defun htmlj-region-is-example (start end)
;  (interactive "*r")
;  (htmlj-replace-kakko start end)
;  (htmlj-kakomu-region start end "PRE")
;  (htmlj-kakomu-region start end "BLOCKQUOTE")
;)

; IMG$B%?%0(B
(defun htmlj-img-tag (str)
  (interactive "sSRC: ")
  (insert (format "<IMG SRC=\"%s\">" str)))

(defun htmlj-get-this-tag ()
  (interactive)
  (save-excursion
     (if (equal "<" (char-to-string (following-char)))
	 (forward-char 1))
    (skip-chars-backward "^<")
    (buffer-substring (point) (progn (skip-chars-forward "^ >") (point)))))

; $B%+!<%=%k0LCV$N%?%0$NCf?H$r<hF@$9$k(B
; htmlj-get-attrib-list$B$+$i8F$V(B
; <$B$N>e$G$bH?1~$9$k$h$&$K$7$?(B 
(defun htmlj-get-this-tag-whole ()
   (interactive)
   (save-excursion
     (if (equal "<" (char-to-string (following-char)))
	 (forward-char 1))
     (skip-chars-backward "^<")
     (buffer-substring (point) (progn (skip-chars-forward "^>") (point)))))

; $B=P8=$9$k%"%H%j%S%e!<%H%j%9%H$r:n$k(B
(defun htmlj-get-attrib-list (str)
  (let ((i 0) (s) (retstr nil) (atr))
    (while (string-match "[^ ][^ ]*" str i)
      (setq i (match-end 0))
      (setq atr (substring str (match-beginning 0) (match-end 0)))
      (string-match "\\([^=][^=]*\\)" atr)
      (setq atr (substring atr (match-beginning 1) (match-end 1)))
      (setq retstr (append retstr (list atr))))
    retstr))

; all-list$B$NMWAG$+$i(Bdel-list$B$NMWAG$r<h$j=|$/(B
; (htmlj-delete-attrib '("HREF" "SRC" "LOW") '("HREF"))  
;      => ("SRC" "LOW")
(defun htmlj-delete-attrib (all-list del-list)
  (while del-list
    (setq all-list (delete (car del-list) all-list))
    (setq del-list (cdr del-list)))
  all-list)

; $B;XDj$5$l$F$$$J$$%"%H%j%S%e!<%H$rBPOCE*$KF~NO$9$k(B
(defun htmlj-supply-attrib ()
  (interactive)
  (save-excursion
    (let ((retstr "") (attrib) (attrib-len) (i 0) (att) (ans))
					; attrib $B$OB0@-%j%9%H(B
      (setq attrib (htmlj-delete-attrib 
		    (cdr (assoc (htmlj-get-this-tag) htmlj-tag-attribute-alist))
		    (htmlj-get-attrib-list (htmlj-get-this-tag-whole))))
      (setq attrib-len (length attrib)) ; $BB0@-$N?t!"%k!<%W@)8f$K;H$&(B
      (while (> attrib-len i)
	; $B3FB0@-$KBP$7!"%_%K%P%C%U%!$GCM$rF~NO$5$;$k(B
	; att: $B8=:_$N%"%H%j%S%e!<%H(B

	(setq att (elt attrib i))
	(setq ans (read-string (format "attribute %s: " att))) 
	(and (not (equal ans ""))
	     (setq retstr (concat retstr (format " %s=\"%s\"" att ans))))
	(setq i (+ i 1)))
	; $B$3$N;~E@$G?7$7$$%Q%i%a!<%?%j%9%H$O(Bretstr$B$KF~$C$F$$$k!#(B
      (skip-chars-forward "^>")
      (insert retstr))
    ))

; $B%+!<%=%k0LCV$K$"$k(BURL$B$K%?%0$r$D$1$k(B
(defun htmlj-put-url-tag-on-string ()
  (interactive)
  (save-excursion
    (let ((sp (point))
	  ep
          (url "")
	  (stag "")
	  (etag "</A>"))
      (setq ep (re-search-forward "\\($\\| \\)")) ; $B9TKv(Bor$B6uGr(B
      (if (not (eq 10 (char-after ep)))
	  (setq ep (- ep 1)))
      ;(message (format "char is %s %d" (char-to-string (char-after ep)) (char-after ep)))
      (setq url (buffer-substring sp ep))
      (setq stag (format "<A HREF=\"%s\">" url))
      (htmlj-kakomu-region-new sp ep stag etag))))

; $B%]%$%s%H$,%?%0$NCf$K$"$k$+$I$&$+$rJV$9(B
;
(defun htmlj-in-a-tag-p ()
  (interactive)
  (save-excursion
     (if (equal "<" (char-to-string (following-char)))
	 t
       (progn (skip-chars-backward "^<>")
	      (if (equal "<" (char-to-string (preceding-char)))
		t
		nil)))))

; $BJD$8%?%0$+!)(B
(defun htmlj-close-tag-p (tag)
  (equal (substring tag 0 1) "/"))

; $BJD$8$F$$$J$$%?%0$rJD$8$k(B
(defun htmlj-insert-closing-tag ()
  (interactive)
  (let ((str (htmlj-close-preceding-tag)))
    (cond ((equal str nil)
	   (message "No tag to insert"))
	  (t (insert (format "</%s>" str))))))

; $B2<@A$1(B of htmlj-insert-closing-tag
(defun htmlj-close-preceding-tag ()
  (interactive)
  (save-excursion
    (let ((curtag nil)
	  (stack ())
	  (loop t)
	  (curpos (point)))
      (if (htmlj-in-a-tag-p) nil
	(progn (while (progn (goto-char (re-search-backward "<[^>]+>"))
			     (setq curtag (htmlj-get-this-tag))
			     (cond ((htmlj-tag-need-single-p curtag) nil)
				   ((htmlj-close-tag-p curtag)
					; $BJD$8%?%0$O(B/$B$r<h$C$F%9%?%C%/$K@Q$`(B
				    (setq stack (cons (substring curtag 1) stack)))
				   ((equal curtag (car stack))
				    (setq stack (cdr stack)))
				   (t (setq loop nil)))
			     loop))
	       curtag)))))



;
; font-lock-mode 
; Time-stamp: <1998/01/31 17:34:48 takuzo>
; - $BBgJ8;z$N$_$H>.J8;z$N$_$r(Bsupport
(defvar htmlj-font-lock-keywords 
  (purecopy
   (list 
    ;; $B$H$j$"$($:(B<>$B$NCf$OA4It(Bkeyword
     '("\\(<[^>]*>\\)" 1 font-lock-keyword-face t)
    ;; comment
    '("\\(<!--\\([^-]\\|-[^-]\\|--[^>]\\)*-->\\)" 1 font-lock-comment-face t)
    ;; DOCTYPE$B$H$+(B
     '("\\(<![a-z]+\\>[^<>]*\\(<[^>]*>[^<>]*\\)*>\\)" 
       1 font-lock-comment-face t)
    ;; bold,italic,underline
    '("<\\(b\\|strong\\)>\\([^<]*\\)</\\1>" 2 bold t)
    '("<\\(address\\|cite\\|em\\|i\\|var\\)>\\([^<]*\\)</\\1>" 2 italic t)
    '("<u>\\([^<]*\\)</u>" 1 underline t)
    ;; $B0z?t$r$H$k$b$N(B
    '("<\\(a\\|form\\|i\\(mg\\|nput\\)\\)\\>\\([^>]*\\)>" 
      3 font-lock-string-face t)
    ;; &keyword
    '("&[^;\n]*;" 0 font-lock-keyword-face t)
    '("=[ \t\n]*\\(\"[^\"]+\"\\)" 1 font-lock-string-face t)

    ;; original$B$N(Bhilit$B$K=`$8$F(B
    '("<\\(h[1-9]\\|title\\).*>\\([^<]*\\)</\\1>" 
      2 font-lock-function-name-face t)
    '("\\(</?p>\\).*" 1 font-lock-function-name-face t)
    '("<a\\s-+href=.*>\\(.*\\)</a>" 1 font-lock-reference-face t)
    ))
  "Expressions and keywords to hilight in font-lock-mode.")

(put 'htmlj-mode 'font-lock-defaults '(htmlj-font-lock-keywords))

